# Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
# SPDX-License-Identifier: MulanPSL-2.0
import os
import argparse
import random
from tqdm import tqdm
import numpy as np
import pandas as pd
from copy import copy
from datetime import datetime
from scipy.spatial.distance import cdist, pdist
import matplotlib.pyplot as plt

SEED = 42
random.seed(SEED)
np.random.seed(SEED)


def mmd(x, y, sigma=None):
    """ Calculate kernel MMD for quantify distribution difference:
    Args:
        x: np.ndarray shape [n, d]
        y: np.ndarray shape [m, d]
        sigma: float, scaling factor
    Returns:
        mmd: mmd values.
    """
    n, d = x.shape
    m, d2 = y.shape
    xy = np.concatenate([x, y], axis=0)

    if sigma is None:
        # Recommend to set the parameter to the median distance between points xy,
        # where xy is the combined sample of x and y.
        # In practice we might take the median of a reasonable subset (e.g. 100 samples).
        _pdists = pdist(xy)
        sigma = np.median(_pdists) / 2

    _cdists = cdist(xy, xy)
    # Just keep the diagonal and everything twice
    # note that sigma should be squared in the RBF to match the Gretton et al heuristic
    k = np.exp(
        (
                -1 / (2 * sigma ** 2)
        ) * _cdists ** 2
    ) + np.eye(n + m) * 1e-5
    k_x = k[:n, :n]
    k_y = k[n:, n:]
    k_xy = k[:n, n:]
    # The diagonals are always 1.
    # note that their code uses the biased (and differently scaled mmd)
    mmd = k_x.sum() / (n * (n - 1)) + k_y.sum() / (m * (m - 1)) - 2 * k_xy.sum() / (n * m)
    return mmd


def mmp_two_sample_test(x, y):
    """ Two sample test with MMD:
        Args:
            x (np.ndarray): data with shape [n, d]
            y (np.ndarray): data with shape [m, d]
        Returns:
            p: P values.
    """
    n, d = x.shape
    m, d2 = y.shape
    xy = np.concatenate([x, y], axis=0)

    _pdists = pdist(xy)
    sigma = np.median(_pdists) / 2
    our_mmd = mmd(x, y, sigma)

    mmds = []
    for i in tqdm(range(1000)):
        xy = xy[np.random.permutation(len(xy))]
        mmds.append(mmd(xy[:n], xy[n:], sigma))

    threshold = np.quantile(mmds, 0.95)  # threshold for 5% significance
    _p = (our_mmd < mmds).mean()
    return _p, threshold


def _loadData(data_path, base=None):
    """Load and prepare data for analyzing

        Args:
            data_path (string): path of numpy data
            base (dict): a dict that stores the baseline configurations and benchmark score
        Return:
            X (numpy array of shape (n,m)): parameter value data,
            y (numpy array of shape (n,)): benchmark scores,
            params (list): a list of parameter names
    """

    # extract parameter names
    knobs = np.load(data_path + '/knobs.pkl',
                    allow_pickle=True)
    # load data
    X = np.load(data_path + '/points.pkl',
                allow_pickle=True)
    # load score
    bench = np.load(data_path + '/bench.pkl',
                    allow_pickle=True)
    # find benchmark results with highest weight as the main objective for analysis
    bench_index = np.argsort([bench[k]['weight'] for k in bench.keys()])[-1]
    score = np.load(data_path + '/score.pkl',
                    allow_pickle=True)
    y = score[:, bench_index]

    if base is None:
        params = [k['name'] for k in knobs]
    else:
        default_params = base["param"]
        x_base = np.array(base["x"]).reshape(1, -1)
        y_base = base["y"]
        params_raw = [k['name'] for k in knobs]
        p_idx = []
        for p1 in default_params:
            for j, p2 in enumerate(params_raw):
                if p1 == p2.partition("@")[0]:
                    p_idx.append(j)
        params = [params_raw[i] for i in p_idx]
        X = X[:, p_idx]
        # concatenate base data with baseline data
        X = np.vstack([x_base, X])
        y = np.hstack([y_base, y])

    return X, y, params


def _segment(X, y, param, figname, n_seg=3):
    """Segment param values that change along with benchmark scores

        Args:
            X (numpy array of shape (n,)): training data, values of a selected parameter named as param
            y (numpy array of shape (n,)): training target for regression,
            param (str): a list of parameter name
            figname (str): path to the figname to save segmentation result
            n_seg (int): number of segments to segment linear regression parameters
        Return:
            mape_error (float): error rate of the regression model
    """
    from sklearn.metrics import mean_absolute_percentage_error as mape
    from sklearn.linear_model import LinearRegression
    from sklearn.gaussian_process import GaussianProcessRegressor, kernels
    from sklearn.tree import DecisionTreeRegressor

    sort_idx = np.argsort(X)
    X = X[sort_idx].reshape(-1, 1)
    y = y[sort_idx].reshape(-1, 1)

    kernel = kernels.Matern()
    # kernel = 1 * RBF(length_scale=1.0, length_scale_bounds=(1e-2, 1e2))
    gpr = GaussianProcessRegressor(kernel=kernel,
                                   n_restarts_optimizer=9,
                                   normalize_y=True)
    gpr.fit(X, y)
    y_pred = gpr.predict(X)
    mape_error = mape(y, y_pred)

    X = X.squeeze()
    y_pred = y_pred.squeeze()

    xs = X + np.random.random(X.shape[0]) * 1e-4
    dys = np.gradient(y_pred, xs)
    idx = np.argwhere(~np.isnan(dys))[:, 0]
    xs = xs[idx]
    ys = y_pred[idx]
    dys = np.gradient(ys, xs)
    dys[np.isnan(dys)] = 0

    plt.figure(figsize=(8, 4))
    rgr = DecisionTreeRegressor(max_leaf_nodes=n_seg)
    rgr.fit(xs.reshape(-1, 1), dys.reshape(-1, 1))

    dys_dt = rgr.predict(xs.reshape(-1, 1)).flatten()

    ys_sl = np.ones(len(xs)) * np.nan

    segs = []
    for dy in np.unique(dys_dt):
        msk = dys_dt == dy
        lin_reg = LinearRegression()
        lin_reg.fit(xs[msk].reshape(-1, 1),
                    ys[msk].reshape(-1, 1))
        ys_sl[msk] = lin_reg.predict(xs[msk].reshape(-1, 1)).flatten()
        segs.append([xs[msk][0], xs[msk][-1]])

        plt.plot([xs[msk][0], xs[msk][-1]],
                 [ys_sl[msk][0], ys_sl[msk][-1]],
                 color='r',
                 zorder=1)

    plt.xlabel(param, fontsize=12)
    plt.ylabel("score", fontsize=12)
    plt.scatter(xs, ys, label='raw data')
    plt.scatter(xs, ys_sl, s=12, label='segmented data', color='k', zorder=5)

    for n in range(len(segs)):
        plt.axvline(x=segs[n][0], linestyle='--', color='m')
        plt.axvline(x=segs[n][-1], linestyle='--', color='m')

    plt.legend(fontsize=12)
    plt.savefig(figname, bbox_inches='tight')

    return mape_error


def _analyze_topK(K, score, data, params, obj):
    """Segment param values that change along with benchmark scores

        Args:
            K (int): Top K
            score (numpy array of shape (n,)): benchmark scores,
            data (numpy array of shape (n,m)): parameter configuration values
            params (list): a list of parameter names
            obj (str): top K for minimum or maximum values
        Return:
            topK_df (dataframe): a dataframe of topK data summary
    """

    from scipy.stats import ks_2samp, binned_statistic

    if K > score.shape[0]:
        K = score.shape[0]

    if obj == "max":
        good_idx = np.argsort(score)[::-1][:K]
        good_data = data[good_idx]
        good_scores = score[good_idx]

        rest_idx = np.argsort(score)[::-1][K:]
        rest_data = data[rest_idx]
    else:
        good_idx = np.argsort(score)[:K]
        good_data = data[good_idx]
        good_scores = score[good_idx]

        rest_idx = np.argsort(score)[K:]
        rest_data = data[rest_idx]

    N = 3 if 3 < K else K
    topK_df = pd.DataFrame()
    for i in range(N):
        topK_dict = {}
        topK_dict["score"] = good_scores[i]
        for p, d in zip(params, good_data[i]):
            topK_dict[p] = d
        topK_df = pd.concat((topK_df,
                             pd.DataFrame(topK_dict, index=[0])))

    p_values = {}
    p_values["score"] = "p_value"
    for i, p in enumerate(params):
        data_bins, data_edges = np.histogram(data[:, i], bins=10)
        good_bin_counts, _, good_bin_id = binned_statistic(
            good_data[:, i],
            np.ones_like(good_data[:, i]),
            statistic="count",
            bins=data_edges)

        rest_bin_counts, _, rest_bin_id = binned_statistic(
            rest_data[:, i],
            np.ones_like(rest_data[:, i]),
            statistic="count",
            bins=data_edges)

        # s1, p1 = ttest_ind(X_pos[:,j], X_neg[:, j], equal_var=False)
        s2, p2 = ks_2samp(good_bin_id,
                          rest_bin_id,
                          alternative='two-sided')  # this is better
        p_values[p] = p2

    topK_df = pd.concat((topK_df,
                         pd.DataFrame(p_values, index=[0])))
    return topK_df


def analysis():
    job_name = "wrk_http_long"
    job_obj = "max"
    K = 5 # top 5 results for summary
    
    data_dir = os.path.join("./data/qe_data", "kafka/kafka_0117_tpe15round_arm")
    result_dir = os.path.join(data_dir, "result")
    data, score, params = _loadData(data_dir)

    # params: by case

    try:
        assert data.shape[0] == score.shape[0]
    except AssertionError:
        l = min([data.shape[0], score.shape[0]])
        data = data[:l]
        score = score[:l]

    non_zero_indice = list(np.where(score != 0)[0])
    data = data[non_zero_indice]
    score = score[non_zero_indice]

    mape_errors = {}
    mape_errors["score"] = "sensi"
    for i, p in enumerate(params):
        X = copy(data[:, i])
        y = copy(score)
        figname = os.path.join(result_dir, f"{job_name}_{p}_segment.png")
        mape_error = _segment(X=X,
                              y=y,
                              param=p,
                              figname=figname,
                              n_seg=4)
        mape_errors[p] = mape_error

    # mape_error_sum = sum(mape_errors.values())
    # for p in params:
    #     mape_errors[p] = mape_errors[p]/mape_error_sum
    mape_error_df = pd.DataFrame(mape_errors, index=[0])

    topK_df = _analyze_topK(K,
                            score,
                            data,
                            params,
                            job_obj)

    result_df = pd.concat((topK_df,
                           mape_error_df))

    result_df.reset_index(inplace=True)
    result_df.drop(columns=["index"], inplace=True)
    result_file = os.path.join(result_dir,
                               f"{job_name}_result.csv")
    result_df.to_csv(result_file, index=False)